/*
 * tmos - spi.c
 *
 * Support for P2 protocol on top of SPI link.
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <tmos/kernel.h>
#include <tmos/spi.h>
#include "p2.h"

/*
 * The P2 protocol uses 16-bit words and we store them, for easier
 * transmit code, in big endian. So during checksuming we must
 * convert endianess.
 */
static void fill_checksum(void __kernel *buf) __reentrant
{
	u16 __kernel *d = buf;
	unsigned char len = ((const u8 __kernel *) buf)[0] & 0xf;
	u16 csum = 0;

	len += 1;
	do {
		csum += be16_to_cpu(*d);
		d++;
		len--;
	} while (len > 0);
	csum = ~csum;
	csum &= 0x7fff;
	*d = cpu_to_be16(csum);
}

/*
 * The P2 protocol uses 16-bit words and we store them, for easier
 * transmit code, in big endian. So during checksuming we must
 * convert endianess. We sum all words including checksum. If checksum
 * is correct we should get 0x7fff.
 */
static char checksum_ok(void __kernel *buf) __reentrant
{
	const u16 __kernel *p = (const u16 __kernel *) buf;
	unsigned char proto_len = ((const u8 __kernel *) buf)[0];
	u16 csum = 0;

	/* check protocol, should be 0x0 */
	if ((proto_len & 0xf0) != 0)
		return 0;

	proto_len += 2;
	csum = 0;
	do {
		csum += be16_to_cpu(*p);
		p++;
		proto_len--;
	} while (proto_len > 0);
	csum &= 0x7fff;
	return csum == 0x7fff;
}

/*
 * This function transmists and recevices P2 protocol frame on SPI
 * interface. This function is not as straightforward as it could
 * be because the spi_send() returns data from previous byte.
 *
 * It's assumed that txlen is non-zero.
 *
 * FIXME: we are probably reading one additional word.
 */
static void __p2_xmit(void __kernel *buf) __reentrant
{
	u8 __kernel *rx = buf;
	u8 __kernel *tx = buf;
	unsigned char txlen;
	unsigned char rxlen;

	spi_start(P2_SYNC);
	spi_send(P2_SYNC);

	txlen = *tx & 0xf;
	txlen++;
	spi_send(*tx);
	tx++;

	*rx = spi_send(*tx);
	rxlen = *rx & 0xf;
	rxlen++;
	if (*rx == 0xff || rxlen < txlen)
		rxlen = 0;
	else
		rxlen -= txlen;

	rx++;
	tx++;

	do {
		*rx = spi_send(*tx);
		rx++;
		tx++;

		*rx = spi_send(*tx);
		rx++;
		tx++;

		txlen--;
	} while (txlen > 0);

	if (!rxlen) {
		*rx = spi_send(P2_STOP);
		spi_send(P2_STOP);
		spi_stop();
	} else {
		do {
			*rx = spi_send(P2_STOP);
			rx++;
			*rx = spi_send(P2_STOP);
			rx++;

			rxlen--;
		} while (rxlen > 0);
		*rx = spi_stop();
	}
}

static char p2_xmit(void __kernel *buf) __reentrant
{
	fill_checksum(buf);
	__p2_xmit(buf);

	return checksum_ok(buf);
}
