/*
 * tmos - prot-v2.c
 *
 * POMMAX2v2 protocol support
 *
 * Copyright (C) 2009..2012  Krzysztof Mazur <krzysiek@podlesie.net>
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/adc.h>
#include <tmos/errno.h>
#include <tmos/reboot.h>
#include <tmos/version.h>
#include "pommax2.h"

#define POMMAX2_VERSION	0x00

#define POMMAX2_PROTO_ANALOGS	0x01
#define POMMAX2_PROTO_CONFIG	0x02

#define POMMAX2_SYNC_0	0xff
#define POMMAX2_SYNC_1	0xff
#define POMMAX2_STOP_0	0x00
#define POMMAX2_STOP_1	0x00

#define POMMAX2_RESET	0x01
#define POMMAX2_READ	0x02
#define POMMAX2_WRITE	0x03

/* registers */
#define POMMAX3_MAJOR	0x80
#define POMMAX3_MINOR	0x81
#define POMMAX3_SAVE	0x82
#define POMMAX3_REBOOT	0x83

#define POMMAX3_REBOOT_MAGIC	0x7eef

#define POMMAX3_FREQ	0x01
#define POMMAX3_FLAGS	0x02
#define POMMAX3_OFFSET	0x03
#define POMMAX3_RMS	0x05
#define POMMAX3_RAW	0x06

static u8 __kernel rx[2 * CHANNELS + 1];
static u8 __kernel *rx_ptr;

static uint16_t csum;

static u16 flags;

static void pommax_send_sample(unsigned char ch) __reentrant
{
	int16_t v;
	uint8_t x;

	adc_wait_ch(ch);
	v = adc_get(ch);

	/*
	 * The POMMAX2v2 protocol allows for 15-bit sample
	 * resolution, but the ADC interface uses 16-bit
	 * format. The least significant bit is just skipped.
	 * If gain correction is enabled we do this in apply_gain();
	 * otherwise we do it here.
	 */
	if (flags & POMMAX3_FLAGS_GAIN)
		v = pommax2_gain_apply(v, ch);
	else
		v >>= 1;

	csum += v;
	x = v >> 8;
	*rx_ptr++ = pommax2_send(x & 0x7f);

	x = v;
	*rx_ptr++ = pommax2_send(x);
}

static void pommax_end_frame(void)
{
	csum = ~csum;
	csum &= 0x7fff;
	pommax2_send(csum >> 8);
	pommax2_send(csum & 0xff);

	pommax2_send(POMMAX2_STOP_0);
	pommax2_send(POMMAX2_STOP_1);
	pommax2_stop();
}

/*
 * This variable holds return value from register read. This value have
 * only 15-bits!
 */
static u16 ret;

static void pommax_read(void)
{
	if (rx[3] < 0x80) {
		ret = pommax2_config_read(rx[3]);
		return;
	}

	switch (rx[3]) {
	case POMMAX3_MAJOR:
		ret = TMOS_VERSION_MAJOR;
		break;
	case POMMAX3_MINOR:
		ret = TMOS_VERSION_MINOR;
		break;
	}
}

static void pommax3_restart(void)
{
	pommax2_start(POMMAX2_SYNC_0);
	pommax2_send(POMMAX2_SYNC_1);
	csum = 0;

	pommax2_send(POMMAX2_VERSION | 1);
	pommax2_send(POMMAX2_PROTO_CONFIG);
	csum += ((POMMAX2_VERSION | 1) << 8) | POMMAX2_PROTO_CONFIG;

	pommax2_send(0);
	pommax2_send(0);
	csum += 0;

	pommax_end_frame();

	machine_restart();
}

static void pommax_write(void)
{
	pommax2_config_write(pack16(rx[5], rx[4]), rx[3]);

	switch (rx[3]) {
	case POMMAX3_FLAGS:
		flags = pack16(rx[5], rx[4]);
		if (flags & POMMAX3_FLAGS_GAIN)
			pommax2_gain_configure();
		break;
	case POMMAX3_FREQ:
		adc_set_freq(pack16(rx[5], rx[4]));
		ret = 0;
		break;
	case POMMAX3_OFFSET:
		pommax2_offset_configure();
		ret = 0;
		break;
	case POMMAX3_RMS:
		pommax2_filter_configure();
		break;
	case POMMAX3_RAW:
		pommax2_filter_configure();
		break;
	case POMMAX3_SAVE:
		if (rx[5] == 0xde && rx[4] == 0xad)
			ret = pommax2_config_save();
		else
			ret = -EINVAL;
		break;
	case POMMAX3_REBOOT:
		ret = -EINVAL;
		if (rx[5] == 0x7e && rx[4] == 0xef)
			pommax3_restart();
	}
}

static u8 old_seq;

static void do_rx_checksum(unsigned char len) __reentrant
{
	const u16 __kernel *p = (const u16 __kernel *) rx;
	const u16 __kernel *end = p + len + 2;
	
	csum = 0;
	do {
		csum += bswap16(*p);
		p++;
	} while (p < end);
	csum &= 0x7fff;
}

static void pommax2_exec_command(void) __reentrant
{
	u8 x;

	/*
	 * process protocol version and message length
	 * we support only protocol version 0 and don't care about
	 * message size
	 */
	x = rx[0];
	if ((x & 0xf0) != 0)
		return;

	do_rx_checksum(x & 0xf);
	if (csum != 0x7fff)
		return;

	/*
	 * we support only configuration protocol
	 */
	x = rx[1];
	if (x != POMMAX2_PROTO_CONFIG)
		return;

	ret = -1;

	/* process command */
	x = rx[2];
	if (x != POMMAX2_RESET) {
		if (((old_seq ^ x) & 0x40) == 0)
			goto out;
	}
	old_seq = x;

	x &= 0x3f;
	ret = 0x0000;
	switch (x) {
	case POMMAX2_READ:
		pommax_read();
		break;
	case POMMAX2_WRITE:
		pommax_write();
		break;
	}

out:
	pommax2_start(POMMAX2_SYNC_0);
	pommax2_send(POMMAX2_SYNC_1);
	csum = 0;

	pommax2_send(POMMAX2_VERSION | 1);
	pommax2_send(POMMAX2_PROTO_CONFIG);
	csum += ((POMMAX2_VERSION | 1) << 8) | POMMAX2_PROTO_CONFIG;

	pommax2_send((ret >> 8) & 0x7f);
	pommax2_send(ret & 0xff);
	csum += ret;

	pommax_end_frame();
}

void pommax2_send_sample(void) __reentrant
{
	unsigned char i;
	u8 x;

	pommax2_start(POMMAX2_SYNC_0);
	pommax2_send(POMMAX2_SYNC_1);

	csum = 0;

	rx_ptr = rx;
	pommax2_send(POMMAX2_VERSION | CHANNELS);
	*rx_ptr++ = pommax2_send(POMMAX2_PROTO_ANALOGS);

	csum += ((POMMAX2_VERSION | CHANNELS) << 8) | POMMAX2_PROTO_ANALOGS;

	for (i = 0; i < CHANNELS; i++)
		pommax_send_sample(i);
	pommax_end_frame();

	pommax2_exec_command();
}

void pommax2_proto_configure(void) __reentrant
{
	flags = pommax2_config_read(POMMAX2_REG_FLAGS);
}
