/*
 * tmos - tboot.c
 * 
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/port.h>
#include <tmos/serial.h>

#ifndef EOF
#define EOF (-1)
#endif

#define LINE_LEN_MAX	48

#if CONFIG_TBOOT_SERIAL_FLOW
#define tboot_xoff() ser_xoff()
#else
#define tboot_xoff() do { } while (0)
#endif

#ifndef ser_flush
#define ser_flush()	do { } while (0)
#endif

#define __line	__highmem

static port_t console;
static char __line line[LINE_LEN_MAX + 1];
static unsigned char line_len;

#if 0
static char keypressed(void)
{
	unsigned char events;
	
	events = __port_poll(console);
	return events & POLLIN;
}
#endif

static void putchar(char c) __reentrant
{
	port_write(console, &c, 1);
}

void puts(const char *s) __reentrant
{
	while (*s)
		putchar(*s++);
}

static int getchar(void) __reentrant
{
	char c;
	msg_ssize_t ret;

	ret = port_read(console, &c, 1);
	if (ret != 1)
		return EOF;
	return c;
}

static void getline(void)
{
	int c;

	line_len = 0;
	line[LINE_LEN_MAX] = '\0';
	for (;;) {
		c = getchar();
		if (c == EOF) {
			line_len = 0;
			return;
		}

		switch ((unsigned char) c) {
		case 127:
		case '\b':
			if (line_len) {
				putchar('\b');
				putchar(' ');
				putchar('\b');
				line_len--;
			}
			break;
		case '\r':
		case '\n':
			tboot_xoff();
			putchar('\n');
			line[line_len] = '\0';
			return;
		default:
			if (line_len < LINE_LEN_MAX) {
				putchar(c);
				line[line_len] = c;
				line_len++;
			}
		}
	}	
}

static char whitespace(char c)
{
	switch (c) {
	case ' ':
	case '\t':
	case '\v':
		return 1;
	};
	return 0;
}

static char __line *stripwhite(char __line *string)
{
	char __line *s;
	char __line *t;

	for (s = string; whitespace(*s); s++)
		;

	if (*s == '\0')
		return s;

	t = s + strlen(s) - 1;
	while ((t > s) && whitespace(*t))
		t--;
	*++t = '\0';

	return s;
}

static char d_to_char(char x) __reentrant
{
	return ((x >= 10) ? 'a' - 10 : '0') + x;
}

static void print_hex(unsigned char val) __reentrant
{
	putchar(d_to_char((val >> 4) & 0xf));
	putchar(d_to_char((val >> 0) & 0xf));
}

void print_whex(unsigned int val) __reentrant
{
	putchar(d_to_char(val >> 12));
	putchar(d_to_char((val >> 8) & 0xf));
	putchar(d_to_char((val >> 4) & 0xf));
	putchar(d_to_char((val >> 0) & 0xf));
}

static void cmd_run(char __line *args) __reentrant
{
	u16 addr = 0;
	void (*prog)(void);

	puts("running program...\n");
	ser_flush();
	local_irq_disable();

	prog = (void * __force)(void __code *__force) addr;
	prog();
}

static void cmd_reboot(char __line *args) __reentrant
{
	u16 addr = CONFIG_CODE_BASE;
	void (*prog)(void);

	puts("reboot...\n");
	ser_flush();

	prog = (void * __force)(void __code *__force) addr;
	prog();
}

static void cmd_hexdump(char __line *args) __reentrant
{
	port_t port;
	off_t off = 0;
	msg_ssize_t ret;
	char buf[16];
	char i;

	if (!*args)
		return;

	port = port_open(args, O_RDONLY);
	if (port == (port_t) -1) {
		puts("hexdump: ");
		puts(args);
		puts(": No such file or directory\n");
		return;
	}
	for (;;) {
		watchdog_refresh();
#if 0
		if (keypressed())
			break;
#endif
		ret = port_pread(port, buf, 16, off);
		if (ret == 0)
			break;
		if (ret < 0) {
			puts("hexdump: ");
			puts(args);
			puts(": read error\n");
			break;
		}
		print_whex(off);
		puts(": ");
		off += ret;
		for (i = 0; i < ret; i++) {
			print_hex(buf[i]);
			putchar(' ');
			if (i == 7)
				putchar(' ');
		}
		putchar('\n');
	}
}

static char __ascii_hex_to_cpu(const char x) __reentrant
{
	if (x >= '0' && x <= '9')
		return x - '0';
	if (x >= 'A' && x <= 'F')
		return x - 'A' + 10;
	return -1;
}

static int ascii_hex_to_cpu(const char __line *x) __reentrant
{
	char hi;
	char lo;
	unsigned char ret;

	hi = __ascii_hex_to_cpu(x[0]);
	if (hi == -1)
		return -1;
	lo = __ascii_hex_to_cpu(x[1]);
	if (lo == -1)
		return -1;
	ret = (unsigned char) lo;
	ret |= (unsigned char) hi << 4;
	return ret;
}

#define IHEX_MAX_LEN	16

static void getline_fast(void)
{
	int c;

	line_len = 0;
	line[LINE_LEN_MAX] = '\0';
	for (;;) {
		c = getchar();
		if (c == EOF) {
			line_len = 0;
			return;
		}

		switch ((unsigned char) c) {
		case '\r':
		case '\n':
			tboot_xoff();
			line[line_len] = '\0';
			return;
		default:
			if (line_len < LINE_LEN_MAX) {
				line[line_len] = c;
				line_len++;
			}
		}
	}	
}

static char format(unsigned char __kernel *buf) __reentrant
{
	int ret;
	unsigned char i;
	unsigned char len = len;

	if (*line != ':')
		return -1;

	i = 0;
	do {
		ret = ascii_hex_to_cpu(line + (unsigned char)(1 + (i << 1)));
		if (ret == -1)
			return ret;

		if (i == 0) {
			if (ret > IHEX_MAX_LEN) {
				puts("record too large\n");
				return 0;
			}
			len = ret + 5;
		}
		*buf = ret;
		i++;
		buf++;
	} while (i < len);
	return 0;
}

static char checksum_bad(unsigned char __kernel *buf) __reentrant
{
	unsigned char csum;
	unsigned char len = buf[0] + 5;

	csum = 0;
	for (len = buf[0] + 5; len > 0; len--, buf++)
		csum += *buf;
	return csum;	
}
	
static char prog_ihex(port_t port) __reentrant
{
	unsigned char buf[IHEX_MAX_LEN + 5];
	msg_ssize_t ret;
	off_t off;

	if (format(buf))
		return 1;

	if (checksum_bad(buf))
		return 1;

	if (buf[3] != 0) {
		if (buf[3] != 1)
			return 1;
		return -1;
	}

	off = pack16(buf[2], buf[1]);
	ret = port_pwrite(port, buf + 4, buf[0], off);
	if (ret != (msg_ssize_t)buf[0])
		return 1;
	return 0;
}

static void cmd_prog(char __line *args) __reentrant
{
	port_t port;
	char ret;
	unsigned char err = 0;

	if (!*args)
		return;

	puts("programming ");
	puts(args);
	puts("...");

	port = port_open(args, O_RDWR);
	if (port == (port_t) -1) {
		puts("prog: ");
		puts(args);
		puts(": No such file or directory\n");
		return;
	}

	for (;;) {
		getline_fast();
		if (*line == '\0')
			break;
		ret = prog_ihex(port);
		if (ret == -1)
			break;
		if (ret)
			err = 1;
	}

	if (err) {
		puts("FAIL\n");
	} else {
		puts("OK\n");
	}
}

struct command {
	const char *name;
	void (*cmd)(char __line *args) __reentrant;
};

static const struct command commands[] = {
	{ "dump", cmd_hexdump },
	{ "prog", cmd_prog },
	{ "run", cmd_run },
	{ "reboot", cmd_reboot },
	{ NULL, NULL },
};

static void execute_line(void)
{
	char __line *cmd;
	char __line *args;
	unsigned char i;
	struct command __code *command;

	cmd = stripwhite(line);
	if (*cmd == '\0')
		return;

	cmd = line;
	i = 0;
	while (line[i] && !whitespace(line[i]))
		i++;

	if (line[i])
		line[i++] = '\0';
	
	while (whitespace(line[i]))
		i++;
	args = line + i;

	for (command = commands; command->name != NULL; command++) {
		if (!strcmp(cmd, command->name)) {
			command->cmd(args);
			return;
		}
	}
	
	puts("unknown command: ");
	puts(cmd);
	putchar('\n');
}
	
static void print_prompt(void)
{
	puts("tboot> ");
}

int app_main(void) __reentrant
{
	console = port_open(CONFIG_CONSOLE_DEVICE, O_RDWR);
	if (console == -1)
		BUG();

	for (;;) {
		print_prompt();
		getline();
		execute_line();
	}
}
