/*
 * wysmod - wysrej.c
 * 
 * Copyright (C) 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/jiffies.h>
#include <tmos/console.h>
#include <tmos/watchdog.h>
#include <tmos/port.h>

#ifndef dir_tx
#define dir_tx()	do { } while (0)
#endif
#ifndef dir_rx
#define dir_rx()	do { } while (0)
#endif

#include "display.h"

static void put_date(char base, const u8 *date)
{
	c_put_hcharz_p(base + 0, date[0]);
	c_putchar_p(base + 2, '-');
	c_put_hcharz_p(base + 3, date[1]);
	c_putchar_p(base + 5, '-');
	c_put_hcharz_p(base + 6, date[2]);
}

static void print_pos(const u8 __kernel *buf)
{
	long pos;

	pos = buf[0] | ((unsigned short) buf[1] << 8)
			| ((long) (char) buf[2] << 16);
	if (pos < 0) {
		pos = -pos;
		c_putchar_p(11, '-');
	} else
		c_putchar_p(11, '+');
	c_put_uintz_e(18, 2, pos);
	c_putchar_p(16, '.');
#ifdef WYSREJ_FULL_POS_RANGE
	c_put_uint_e(15, 4, pos / 100);
#else
	c_put_uint_e(15, 4, (unsigned int)(pos >> 2) / 25);
#endif
	c_putchar_p(19, 'm');
}

/*
 * LCD
 *	 00000000001111111111
 * 	 01234567890123456789
 * 	"V=XX.Xm/s S=XXXX.XXm"
 * 	"XX-XX-XX    XX-XX-XX"
 */

void wysrej(void)
{
	static __code u8 vel_cmd[] = { 0xff, 0x0f, 0x10, 0x00 };
	static __code u8 pos_cmd[] = { 0xff, 0x0f, 0x11, 0x00 };
	static __code u8 data_range_cmd[] = { 0xff, 0x0f, 0x12, 0x00 };
	u8 buf[12];
	char ret;
	int vel;
	unsigned char retry = 5;
	unsigned char ar3c_resp = 5;

	for (;;) {
		watchdog_refresh();

		port_read_timeout(wysmod_fd, buf, sizeof(buf), MS_JIFFIES(1));

		dir_tx();
		tmos_memcpy(buf, vel_cmd, sizeof(vel_cmd));
		port_write(wysmod_fd, buf, sizeof(vel_cmd));
		dir_rx();

		ret = port_read_timeout(wysmod_fd, buf, 2, MS_JIFFIES(200));
		if (ret != 2) {
			retry--;
			if (!retry)
				break;
			continue;
		}
		ret = port_read_timeout(wysmod_fd, buf, sizeof(buf),
				MS_JIFFIES(50));
		if (ret > 0) {
			retry--;
			if (!retry)
				break;
			continue;
		}
		retry = 5;
		vel = buf[0] + ((unsigned short) buf[1] << 8);

		dir_tx();
		tmos_memcpy(buf, pos_cmd, sizeof(pos_cmd));
		port_write(wysmod_fd, buf, sizeof(pos_cmd));
		dir_rx();

		ret = port_read_timeout(wysmod_fd, buf, 3, MS_JIFFIES(200));
		if (ret != 3)
			continue;

		c_putchar_p(0, 'V');
		if (vel < 0) {
			vel = -vel;
			c_putchar_p(1, '-');
		} else
			c_putchar_p(1, '+');

		c_putchar_p(4, '.');
		c_putchar_p(5, d_to_char(vel % 10));
		c_put_uint_e(3, 2, vel / 10);
		c_puts_p(6, "m/s S");

		print_pos(buf);
		c_flush();

		watchdog_refresh();

		dir_tx();
		tmos_memcpy(buf, data_range_cmd, sizeof(data_range_cmd));
		port_write(wysmod_fd, buf, sizeof(data_range_cmd));
		dir_rx();

		ret = port_read_timeout(wysmod_fd, buf, 12, MS_JIFFIES(200));
		watchdog_refresh();
		if (ret == 10) {
			ar3c_resp = 5;
			c_puts_p(LCD_COLS + 8, "    ");
			put_date(LCD_COLS + 0, buf);
			put_date(LCD_COLS + 12, buf + 5);
		} else if (ret == 12) {
			/*
			 * if we're connected to AR-3c switch to WYDAL
			 * mode
			 */
			ar3c_resp--;
			if (!ar3c_resp)
				break;
			c_puts_p(LCD_COLS + 8, "    ");
			put_date(LCD_COLS + 0, buf + 1);
			put_date(LCD_COLS + 12, buf + 7);
		} else {
			c_puts_p(LCD_COLS, "--.--.--    --.--.--");
		}
		c_flush();
	}
}
