/*
 * tmos - page.c
 *
 * Page allocator
 *
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <limits.h>

/*
 * We use simple stack of free pages. We keep pointer to the first page.
 * The pointer to the next page is stored at the beginning of free page.
 *
 * The page list is protected by disabling preemption - the page
 * allocator cannot be used in any interrupt context.
 */
static unsigned char __xdata first_page;

/**
 * page_alloc - allocate page
 *
 * This function returns page identifier; PAGE_NULL on error.
 */
page_t page_alloc(void) __reentrant
{
	page_t page;

	preempt_disable();
	page = first_page;
	if (page != PAGE_NULL) {
		page_t __xdata *p;

		p = page_to_virt(page);
		first_page = *p;
	}
	preempt_enable();
	return page;
}

/**
 * page_free - free page allocated by alloc_page()
 * @page: page identifier
 */
void page_free(page_t page) __reentrant
{
	page_t __xdata *p;

	preempt_disable();
	p = page_to_virt(page);
	*p = first_page;
	first_page = page;
	preempt_enable();
}

void page_alloc_init(void)
{
	page_t i = NUM_PAGES - 1;

	first_page = PAGE_NULL;
	do {
		page_free(i);
		i--;
	} while (i > 0);
}
