/*
 * tmos - softirq.c
 * Copyright (C) 2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <tmos/kernel.h>
#include <asm/8051.h>
#include <asm/softirq.h>
#include <asm/irqs.h>
#include <tmos/sched.h>
#include <tmos/spruked.h>

extern struct context __kernel *current;

/*
 * there can be only one active softirq
 */
static __bit softirq_running = 0;

__bit __softirq_pending;

/**
 * softirq_return - return from softirq context
 *
 * The softirq code should jump to this function to exit softirq context.
 */
static void softirq_return(void) __using(SOFTIRQ_BANK) __naked
{
	setup_registers(SOFTIRQ_BANK);

#ifdef __INLINE_ASM
	__asm
		pop	_bp
		pop	b
		pop	acc
		pop	psw

		clr	_softirq_running

		/*
		 * The interrupts must be enabled just before ret to
		 * use interrupt shadow to prevent possible stack overflow
		 * when interrupt executes between clearing softirq_running
		 * and returning.
		 */
		setb	_EA
		ret
	__endasm;
#endif
}

/**
 * softirq - the main softirq function
 *
 * The interrupts that want to run softirqs should return to this function.
 */
void softirq(void) __using(SOFTIRQ_BANK) __naked
{
	/* change register bank to 1 */
	__asm__(mov psw,#(SOFTIRQ_BANK << 3));
	setup_registers(SOFTIRQ_BANK);
restart:
	__asm__(push dpl);
	__asm__(push dph);

	do {
		__softirq_pending = 0;
		if (timers_pending()) {
			clear_timers_pending();
			do_timers();
		}
		call_spruked_softirq();
	} while (__softirq_pending);
	__asm__(pop dph);
	__asm__(pop dpl);
	irq_switch_task();

	/*
	 * Check the pending again with interrupts disabled, so we will
	 * see every pending bit set by an interrupt that saw
	 * softirq_running=1.
	 */
	local_irq_disable();
	if (__softirq_pending)
		goto restart;
	__asm__(ljmp	_softirq_return);
}

#ifdef CONFIG_MULTITASKING
/**
 * thread_switch - switch current thread
 */
void thread_switch(void) __naked
{
	setup_registers(0);
	softirq_running = 1;

#ifdef __INLINE_ASM
	__asm
		push psw
		push acc
		push b
		push _bp
		setb EA
	__endasm;
#endif

	/* change register bank to 1 */
	__asm__(mov psw,#(SOFTIRQ_BANK << 3));
	setup_registers(SOFTIRQ_BANK);

	/* switch current task */
	switch_task();

	/* return to kernel context */
	__asm__(ljmp _softirq_return);
}
#endif

/**
 * __irq_exit - exit from low priority interrupt helper
 *
 * This function is be called by low priority interrupt that
 * triggers softirq. If the softirq is already running this
 * function just returns from interrupt.
 */
void __irq_exit(void) __using(LOW_PRIORITY_BANK) __naked
{
	setup_registers(LOW_PRIORITY_BANK);

	if (!__softirq_pending || softirq_running) {
		/*
		 * softirq is already running. We go to softirq
		 * be just returing from interrupt.
		 */
#ifdef __INLINE_ASM
		__asm
			pop	_bp
			pop	b
			pop	acc
			pop	psw
			reti
		__endasm;
#endif
	}

	softirq_running = 1;
#ifdef __INLINE_ASM
	/*
	 * setup stack for return to softirq() function and return from
	 * interrupt
	 */
	__asm
		mov	r0, #_softirq
		push	ar0
		mov	r0, #(_softirq >> 8)
		push	ar0
		reti
	__endasm;
#endif
}

/**
 * __irq_simple_exit - exit from a simple, low priority interrupt helper
 *
 * This function is be called by a simple, low priority interrupt that
 * triggers softirq. If the softirq is already running this
 * function just returns from interrupt.
 */
void __irq_simple_exit(void) __using(LOW_PRIORITY_BANK) __naked
{
	setup_registers(LOW_PRIORITY_BANK);

	if (!__softirq_pending || softirq_running) {
		/*
		 * softirq is already running. We go to softirq
		 * be just returing from interrupt.
		 */
#ifdef __INLINE_ASM
		__asm
			pop	acc
			pop	psw
			reti
		__endasm;
#endif
	}

	softirq_running = 1;
#ifdef __INLINE_ASM
	/*
	 * setup stack for return to softirq() function and return from
	 * interrupt
	 */
	__asm
		push	b
		push	_bp
		mov	r0, #_softirq
		push	ar0
		mov	r0, #(_softirq >> 8)
		push	ar0
		reti
	__endasm;
#endif
}

void call_softirq(void) __naked
{
	softirq_running = 1;
	barrier();

#ifdef __INLINE_ASM
	__asm
		push psw
		push acc
		push b
		push _bp
	__endasm;
#endif

	__asm__(ljmp _softirq);
}
