/*
 * tmos - pommax2.h
 *
 * POMMAX2 digital filters
 *
 * Copyright (C) 2010, 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

struct pommax2_lowpass_state {
	s16 y1;
	s16 y2;
	s16 pad;
	s8 e;
	s8 e1;
};

struct pommax2_rms_state {
	u16 y3;
	u16 y1;
	u16 y2;
	u8 e1;
	u8 e2;
};

union pommax2_state {
	struct pommax2_lowpass_state lowpass;
	struct pommax2_rms_state rms;
};

static union pommax2_state pommax2_state;

extern long ar2_y;
extern int ar2_t;

#define y ar2_y
#define t ar2_t
#define _y _ar2_y
#define _t _ar2_t

#define l_y1 pommax2_state.lowpass.y1
#define l_y2 pommax2_state.lowpass.y2
#define l_e pommax2_state.lowpass.e
#define l_e1 pommax2_state.lowpass.e1

#define r_y1 pommax2_state.rms.y1
#define r_y2 pommax2_state.rms.y2
#define r_y3 pommax2_state.rms.y3
#define r_e1 pommax2_state.rms.e1
#define r_e2 pommax2_state.rms.e2

static void lowpass_stage1(void) PROCESS_ATTR;

static void pommax2_raw(s16 x) PROCESS_ATTR
{
	r_y3 = r_y2 = r_y1 = x << 3;
	r_e1 = r_e2 = 0;
}

static void pommax2_lowpass(s16 x) PROCESS_ATTR
{
	register int t;

	lowpass_stage1();

	/*
	 * Both l_y1 is 15-bit signed value and PARM_B * x are 15-bit
	 * signed values and we first substract those values
	 * and later add to y to save one 32-bit operation.
	 */
	t = PARM_B * x - l_y2;
	y += t;

	/* error feedback */
	t = l_e << 1;
	t = -t + l_e1;
	y = y + t;

	l_e1 = l_e;
	l_e = (int8_t)(-((uint8_t)y));

	/* division */
	y += 127;
	t = y >> 8;

	/* integrate */
	t = l_y1 + t;

	if (t > 16383)
		t = 16383;
	else if (t < -16384)
		t = -16384;
	l_y2 = l_y1;
	l_y1 = t;
}

/*
 * RMS with 0.5% accuracy
 *
 * Performance:
 * Resolution:		8-bit
 * Frequency band:	13 Hz to 800 Hz
 * Group delay:		160 ms
 * Rise to 95% of RMS	330 ms
 * Rise to +/- 0.5%	560 ms
 *
 * Pass-band:
 * 	0.04 dB		0.13 Hz
 * 	0.5 dB		0.48 Hz
 * 	3 dB		1.28 Hz
 * Stop-band:
 * 	40 dB		19.8 Hz
 * 	48 dB		31.4 Hz
 *
 * Accuracy:
 * 				5 Hz	10 Hz	>= 13 Hz	50 Hz
 * Max error			1.48%	0.54%	0.39%		0.39%
 * 2sigma			1.26%	0.42%	0.32%		0.29%
 */
static void pommax2_rms(s16 x) PROCESS_ATTR
{
	register int t;
	register unsigned char x8;

	if (x == -2048)
		x = 2047;
	else if (x < 0)
		x = -x;
	x8 = x >> 3;

	t = ((unsigned int)(x8 * x8 + 1) >> 1) - r_y1 + r_e1;
	r_y1 += t >> 7;
	r_e1 = t & 0x7f;

	t = r_y1 - r_y2 + r_e2;
	r_y2 += t >> 6;
	r_e2 = t & 0x3f;

	x8 = (unsigned int) r_y3 >> 8;
	t = (int)r_y2 - (int)((unsigned int)(x8 * x8 + 1) >> 1);
	r_y3 += t;
	if (t >= 0 && r_y3 < (unsigned int)t)
		r_y3 = 65535;
}

static void lowpass_stage1(void) PROCESS_ATTR
{
#if 1
	__bit x;

	t = l_y2 - l_y1;
	x = 0;
	if (t < 0) {
		t = -t;
		x = 1;
	}

	B = PARM_A;	/* 3 clocks */
	/* 42 clocks */
#ifdef __INLINE_ASM
	__asm
		mov	a, (_t + 0)		; 2
		mul	ab			; 9
		mov	(_y + 0), a		; 2
		mov	r4, b			; 2
	__endasm;
#endif

	B = PARM_A;	/* 3 clocks */
#ifdef __INLINE_ASM
	__asm
		mov	a, (_t + 1)		; 2
		mul	ab			; 9
		add	a, r4			; 1
		mov	(_y + 1), a		; 2
		clr	a			; 1
		mov	(_y + 3), a		; 2
		addc	a, b			; 2
		mov	(_y + 2), a		; 2
	__endasm;
#endif
	if (!x)
		y = -y;
#else
	y = 232L * t;
#endif
}
